;; Generated package description from corfu.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "corfu" "2.15.0.20260905.2" "COmpletion in Region FUnction" '((emacs "29.1") (compat "31")) :commit "5869254349a035e16d656eb232a52cf0e163d531" :authors '(("Daniel Mendler" . "mail@daniel-mendler.de")) :maintainer '("Daniel Mendler" . "mail@daniel-mendler.de") :keywords '("abbrev" "convenience" "matching" "completion" "text") :url "https://github.com/minad/corfu")
